/***************************************************************************
 *                                                                         *
 *   GALFIL.C                                                              *
 *                                                                         *
 *   Copyright (c) 1993-1995 GALACTICOMM, Inc.    All Rights Reserved.     *
 *                                                                         *
 *   File Libraries - Main Routines                                        *
 *                                                                         *
 *                                                 - D. Pitchford  9/28/93 *
 *                                                                         *
 ***************************************************************************/

#include "gcomm.h"
#include "majorbbs.h"
#include "gcspsrv.h"
#include "fcntl.h"
#include "galfil.h"
#include "filexfer.h"
#include "fsdbbs.h"
#include "reserve.h"
#include "galfilh.h"

#define FILREV "$Revision:   1.0.1.1.1.0  $"

static
struct module flmod={              /* module interface block               */
     "",                           /*    name used to refer to this module */
     lslon,                        /*    user logon supplemental routine   */
     lsinp,                        /*    input routine if selected         */
     lssta,                        /*    status-input routine if selected  */
     NULL,                         /*    "injoth" routine for this module  */
     NULL,                         /*    user logoff supplemental routine  */
     lshang,                       /*    hangup (lost carrier) routine     */
     lsclean,                      /*    midnight cleanup routine          */
     lsdel,                        /*    delete-account routine            */
     lsend                         /*    system shutdown routine           */
};

struct fluser *flu,                /* current user pointed to (disk)       */
              *fluarr;             /* memory array                         */

struct fluson *flo;                /* online user pointed to               */

struct flusave flusmem;            /* memory for loading/saving fluser     */

struct fllib *curlib,              /* current Library in use               */
             **alphlibs=NULL;      /* alphabetical pointers to Libraries   */

struct libdisk *libupd,            /* memory for loading/saving Libraries  */
               *libedt;            /* memory for editing Libraries         */

struct taglib *ftag;               /* current tagged file                  */

struct modeview mvarr[NVIEWS];     /* custom viewing routines              */

struct key0 countkey;              /* key for counting uploads in Libraries*/

struct key7 *kwbuf;                /* buffer for keyword insertion         */

int srcpool;                       /* handle to pool of search structs     */
int *srcharr;                      /* array of search handles, one per user*/
struct srcs *srcu;                 /* ptr to cur (setsrc()'d) search struct*/

union srch srcactive;              /* active, found record for speed-search*/

int curfusr;                       /* cur (fluoff()'d) file library user   */

FILE *flmsg,                       /* pointer for .msg file                */
     *csrc,                        /* pointer for general copying of files */
     *cdst,                        /* pointer for general copying of files */
     *movs,                        /* pointer for sysop copying of files   */
     *movd;                        /* pointer for sysop copying of files   */

BTVFILE *flldat,                   /* data file for Libraries              */
        *flfdat,                   /* data file for files                  */
        *flkdat;                   /* data file for keywords               */

char *arrower,                     /* cursor key detection flag            */
     *tempdir,                     /* default user upload temp directory   */
     *nochgkey,                    /* key to never be charged to d/l       */
     *listkey,                     /* key to be able to d/l file lists     */
     *nonkwds,                     /* non-keywords                         */
     *flsysop,                     /* key required to be Library sysop     */
     *defplop,                     /* default primary Lib-Op               */
     *copydir,                     /* destination directory, copied files  */
     *defdesc,                     /* default description for a Library    */
     *deflname,                    /* name of default Library              */
     *descedit,                    /* memory for file descriptions         */
     *dargv[MAXKWDS],              /* points to defdesc, as margv to input */
     *defproto,                    /* default d/l protocol for all users   */
     flmisc[MISCSIZE],             /* array for misc. test variable        */
     movflib[FLNAMESZ],            /* array for moving files from Library  */
     movtlib[FLNAMESZ],            /* array for moving files to Library    */
     movfnam[FLFILENM],            /* array for moving files file name     */
     countlib[FLNAMESZ],           /* Library currently being counted      */
     tagtmpp[TSHLEN+1],            /* memory for temporary wild file path  */
     tagtmpf[FLFILENM];            /* memory for temporary wild file name  */

void *flomem,                      /* memory for users online              */
     *libmem,                      /* memory for Libraries                 */
     *libmem2,                     /* more memory for adding Libraries     */
     *ftagmem;                     /* memory for tagged files              */

int numlib,                        /* number of libs memory allocated for  */
    numladd,                       /* max. number of libs to possibly add  */
    flstt,                         /* state for the File Libraries         */
    longsrch,                      /* require brute force searches         */
    vdd,                           /* half of the vda declaration          */
    freeuls,                       /* if true, never charge while uploading*/
    dargc,                         /* count of parsed words by darsdesc()  */
    notifsys,                      /* if true, notify sysops of unapp files*/
    notiflop,                      /* if true, notify libops of unapp files*/
    numoflib,                      /* actual number of Libraries           */
    comeff,                        /* estimated communications efficiency  */
    ncatlibs,                      /* actual number of top-level Libraries */
    auditall,                      /* if true, audit all file transfers    */
    unifile,                       /* if true, require unique file names   */
    pfnceil,                       /* profanity ceiling for File Libraries */
    chgabt,                        /* if true, charge for aborted downloads*/
    defallib,                      /* search all Libraries from d/l prompt */
    systags,                       /* number of ftagmem to allocate        */
    defstyp,                       /* default search/list mode             */
    defslib,                       /* default search/list range, one/all   */
    stplist,                       /* strip ansi always in file listings   */
    countuls[ULDAYS],              /* counter of Library files added daily */
    clnlevel,                      /* level of cleanup to run              */
    kwaut,                         /* flag, keywords being written now     */
    maxjoin,                       /* # of joined libs max for fast search */
    srcindex,                      /* index for found record in speed-srch */
    fltagf;                        /* is current file tag a Library tag    */

long byteroom,                     /* bytes free for uploading to a Library*/
     thresh;                       /* bytes assured free, copydir's drive  */

struct reserve *librsv=NULL;       /* File Libraries reservations domain   */

char *libedits="\
LIBDATE \
LIBTIME \
LIBNAME \
LIBDOS \
LIBDESC \
LIBDESC0 \
LIBDESC1 \
LIBDESC2 \
LIBDESC3 \
LIBDESC4 \
LIBPATH \
LIBHID(ALT=YES ALT=NO MULTICHOICE) \
LIBCBD(ALT=YES ALT=NO MULTICHOICE) \
LIBRDO(ALT=YES ALT=NO MULTICHOICE) \
LIBAUL(ALT=YES ALT=NO MULTICHOICE) \
LIBADL(ALT=YES ALT=NO MULTICHOICE) \
LIBFUL(ALT=YES ALT=NO MULTICHOICE) \
LIBFDL(ALT=YES ALT=NO MULTICHOICE) \
LIBCDF \
LIBCDK \
LIBROY \
LIBMFL \
LIBMBT \
LIBMUP \
LIBVIS \
LIBDLD \
LIBULD \
LIBOVW \
LIBAPU \
LIBLOP \
LIBPLOP \
DONE(ALT=SAVE ALT=EDIT ALT=QUIT MULTICHOICE)";

char *filedits="\
FILNAME \
FILLIB \
FILDATE \
FILTIME \
FILSIZE \
FILDL \
FILULD \
FILULT \
FILDLT \
FILCHG \
FILUSER \
FILTXT00 \
FILTXT01 \
FILTXT02 \
FILTXT03 \
FILTXT04 \
FILTXT05 \
FILTXT06 \
FILTXT07 \
FILTXT08 \
FILTXT09 \
FILTXT10 \
FILTXT11 \
FILTXT12 \
FILTXT13 \
FILTXT14 \
DONE(ALT=SAVE ALT=EDIT ALT=QUIT MULTICHOICE)";

char *libeditf="\
LIBDATE=%s%c\
LIBTIME=%s%c\
LIBNAME=%s%c\
LIBDOS=%s%c\
LIBDESC=%s%c\
LIBDESC0=%s%c\
LIBDESC1=%s%c\
LIBDESC2=%s%c\
LIBDESC3=%s%c\
LIBDESC4=%s%c\
LIBPATH=%s%c\
LIBHID=%s%c\
LIBCBD=%s%c\
LIBRDO=%s%c\
LIBAUL=%s%c\
LIBADL=%s%c\
LIBFUL=%s%c\
LIBFDL=%s%c\
LIBCDF=%ld%c\
LIBCDK=%ld%c\
LIBROY=%d%c\
LIBMFL=%ld%c\
LIBMBT=%ld%c\
LIBMUP=%ld%c\
LIBVIS=%s%c\
LIBDLD=%s%c\
LIBULD=%s%c\
LIBOVW=%s%c\
LIBAPU=%s%c\
LIBLOP=%s%c\
LIBPLOP=%s%c";

char *fileditf="\
FILNAME=%s%c\
FILLIB=%s%c\
FILDATE=%s%c\
FILTIME=%s%c\
FILSIZE=%s%c\
FILDL=%s%c\
FILULD=%s%c\
FILULT=%s%c\
FILDLT=%s%c\
FILCHG=%s%c\
FILUSER=%s%c\
FILTXT00=%s%c\
FILTXT01=%s%c\
FILTXT02=%s%c\
FILTXT03=%s%c\
FILTXT04=%s%c\
FILTXT05=%s%c\
FILTXT06=%s%c\
FILTXT07=%s%c\
FILTXT08=%s%c\
FILTXT09=%s%c\
FILTXT10=%s%c\
FILTXT11=%s%c\
FILTXT12=%s%c\
FILTXT13=%s%c\
FILTXT14=%s%c";

void EXPORT
init__galfil(void)                 /* system initialization routine        */
{
     int ncpers,ninmem,i;
     unsigned areasiz;

     stzcpy(flmod.descrp,gmdnam("GALFIL.MDF"),MNMSIZ);
     flstt=register_module(&flmod);
     flmsg=opnmsg("galfil.mcv");
     iniopts();
     flfdat=opnbtv("galfilf.dat",FLFILREC+DESCSIZ);
     flldat=opnbtv("galfill.dat",sizeof(struct libdisk));
     numlib=(int)cntrbtv()+5;
     numladd=numlib+20;
     if (!longsrch) {
          flkdat=opnbtv("galfilk.dat",FLKEYREC);
     }
     if (tempdir[0] == '\0') {
          tempdir="LIBTEMP";
     }
     if (copydir[0] == '\0' || weirddir(copydir)) {
          copydir="LIBCOPY";
     }
     arrower=(char *)alczer(nterms);
     descedit=(char *)alczer(DESCSIZ);
     libedt=(struct libdisk *)alczer(sizeof(struct libdisk));
     libupd=(struct libdisk *)alczer(sizeof(struct libdisk));
     fluarr=(struct fluser *)alczer(nterms*sizeof(struct fluser));
     areasiz=sizeof(struct srcs)-sizeof(char);
     areasiz+=maxjoin*sizeof(union srch)*NUINRS;
     ncpers=numopt(NCPERS,1,150);
     ninmem=max(2,(nterms-1)/ncpers+1);
     srcpool=newpool(areasiz,nterms,ninmem);
     srcharr=(int *)alcmem(nterms*sizeof(int));
     for (i=0 ; i < nterms ; i++) {
          srcharr[i]=rsvarea(srcpool);
          if (srcharr[i] == -1) {
               catastro("CAN'T RESERVE LIBRARY SEARCH MEMORY AREA!");
          }
     }
     if (!longsrch) {
          kwbuf=(struct key7 *)alczer(SIZKWBUF*sizeof(struct key7));
     }
     kwaut=0;
     ftagmem=alcblok(systags,sizeof(struct taglib));
     flomem=alcblok(nterms,sizeof(struct fluson));
     libmem=alcblok(numlib,sizeof(struct fllib));
     libmem2=NULL;
     setmem(mvarr,NVIEWS*sizeof(struct modeview),0);
     register_modeview("ZIP",CMPVIEW,NULL);
     register_modeview("ARC",CMPVIEW,NULL);
     register_modeview("LZH",CMPVIEW,NULL);
     register_modeview("ZOO",CMPVIEW,NULL);
     register_modeview("ICE",CMPVIEW,NULL);
     register_modeview("PSE",CMPVIEW,NULL);
     register_modeview("ZSE",CMPVIEW,NULL);
     register_modeview("TXT",LSTVIEW,NULL);
     register_modeview("BAT",LSTVIEW,NULL);
     register_modeview("INI",LSTVIEW,NULL);
     loadlibs();
     init_galfiluz();
     vdd=max(MAXKWDS*FLKEYSIZ,fsdroom(LIBEDITN,libedits,0));
     vdd=max(vdd,fsdroom(LIBEDITA,libedits,0));
     vdd=max(vdd,fsdroom(FILEDITN,filedits,0));
     vdd=max(vdd,fsdroom(FILEDITA,filedits,0));
     dclvda(2*vdd);
     if (sameto("_LIBOP",flsysop)) {
          shocst("GALFIL: WARNING","FLSYSOP key can't be _LIBOP pseudokey");
     }
     else {
          register_pseudok("_LIBOP",psislibop);
     }
     mkdir(tempdir);
     mkdir(copydir);
     purgecpy(0);
     regtvars();
     rtkick(5,libkick);
     initlibrsv();
     init_cslib();
}

void
initlibrsv(void)                   /* init File Library reservations domain*/
{
     if (librsv == NULL) {
          librsv=inireserve(nterms,&usrnum,3,FLNAMESZ,FLFILENM,amrlib);
     }
}

void
iniopts(void)                      /* read in configuration options        */
{
     longsrch=ynopt(LONGSRCH);
     pfnceil=numopt(PFNCEIL,0,3);
     flsysop=stgopt(FLSYSOP);
     nonkwds=stgopt(NONKWDS2);
     tempdir=checkdir(pthopt(TEMPDIR));
     comeff=numopt(COMEFF,1,100);
     thresh=lngopt(THRESH,1,1000)*1024L*1024L;
     auditall=ynopt(AUDITALL);
     chgabt=ynopt(CHGABT);
     defallib=ynopt(DEFALLIB);
     nochgkey=stgopt(NOCHGKEY);
     defdesc=stgopt(DEFDESC);
     deflname=stgopt(DEFLNAME);
     defplop=stgopt(DEFPLOP);
     listkey=stgopt(LISTKEY);
     notifsys=ynopt(NOTIFSYS);
     notiflop=ynopt(NOTIFLOP);
     unifile=ynopt(UNIFILE);
     freeuls=ynopt(FREEULS);
     stplist=ynopt(STPLIST);
     systags=numopt(SYSTAGS,20,3000);
     copydir=checkdir(pthopt(COPYDIR));
     defproto=stgopt(DEFPROTO);
     clnlevel=numopt(CLNLEVEL,0,2);
     switch (chropt(DEFSTYP)) {
     case 'd':
     case 'D':
          defstyp=2;
          break;
     case 'f':
     case 'F':
          defstyp=3;
          break;
     case 'k':
     case 'K':
          defstyp=4;
          break;
     case 'n':
     case 'N':
          defstyp=5;
          break;
     case 'w':
     case 'W':
          defstyp=6;
          break;
     case 'q':
     case 'Q':
          defstyp=7;
          break;
     default:
          defstyp=1;
     }
     defslib=tokopt(DEFSLIB,"ASK","CURLIB","ALL",NULL);
     if (defslib == 0) {
          defslib++;
     }
     maxjoin=numopt(MAXJOIN,5,500)+1;   /* include the top-level Library   */
}

int
psislibop(                         /* pseudokey: is user a libop?          */
int unum,                          /* test for _LIBOP! to determine sysop  */
char *lock)
{
     if (!sameto("_LIBOP",lock)) {
          return(0);
     }
     fluoff(unum);
     if (curlib == NULL) {
          return(0);
     }
     if (sameto("_LIBOP",curlib->libop)) {
          shocst("GALFIL: WARNING",
                 "LIBOP key can't be _LIBOP in Library %s",curlib->libname);
          return(0);
     }
     if (lock[strlen(lock)-1] == '!') {
          return(haskey(flsysop));
     }
     return(isflop(curlib));
}

char *
cnclib(void)                       /* grab a Library name                  */
{
     static char retlib[FLNAMESZ];
     char *rp;

     setmem(rp=retlib,FLNAMESZ,0);
     while (*nxtcmd != '\0' && *nxtcmd != ' ' && rp-retlib < FLNAMESZ-1) {
          *rp++=*nxtcmd++;
     }
     return(strupr(retlib));
}

int
isflop(                            /* is current user an operator at all?  */
struct fllib *libptr)
{
     if (libptr == NULL) {
          return(0);
     }
     return(haskey(flsysop)
         || libptr->libop[0] != '\0' && haskey(libptr->libop)
         || stricmp(usaptr->userid,libptr->primary) == 0);
}

int
scanlib(                           /* count up total files in Library      */
int user,                               /* ... 1=use user's access keys    */
struct fllib *libptr)
{
     int loop,loop2;
     struct fllib *lib2;

     libptr->libs=0;
     libptr->totfiles=libptr->numfiles-libptr->appwait;
     for (loop=0 ; loop < numoflib ; loop++) {
          lib2=liboff(loop);
          if (lib2->cat[0][0] != '\0') {
               for (loop2=0 ; loop2 < FLNMCATS ; loop2++) {
                    if (sameas(lib2->cat[loop2],libptr->libname)
                       && (!user || haslibkey(lib2,lib2->keyreq))) {
                         libptr->libs++;
                         if (lib2 != libptr) {
                              libptr->totfiles+=lib2->numfiles-lib2->appwait;
                         }
                    }
               }
          }
     }
     return(libptr->libs != 0);
}

int
wldgetag(void)                     /* get next tagged wild file            */
{
     struct fllib *libptr;

     if (readline(flo->fsrc)) {
          inplen=strlen(input);
          parsin();
          if (margc >= 2 && (libptr=libfind(margv[0])) != NULL) {
               setbtv(flfdat);
               if ((libptr->flags&FLGDOS)
                  || acqbtv(NULL,compkey(libptr->libname,margv[1]),COMPLF)) {
                    sprintf(tagtmpp,"%s\\%s",libpath(libptr),margv[1]);
                    if (fnd1st(&flo->fb,tagtmpp,0)) {
                         stzcpy(tagtmpf,flo->fb.name,FLFILENM);
                         strcpy(spc->libname,libptr->libname);
                         spc->size=flo->fb.size;
                         return(1);
                    }
               }
          }
     }
     fclose(flo->fsrc);
     flo->fsrc=NULL;
     return(0);
}

int
dlbgnwld(void)                     /* begin download of wild tags          */
{
     int tagnum;

     for (tagnum=0 ; tagnum < systags ; tagnum++) {
          if (tagoff(tagnum) == spc) {
               break;
          }
     }
     flo->fsrc=fopen(spr("%s\\TAGS%d.fil",copydir,tagnum),FOPRB);
     return(flo->fsrc == NULL ? 0 : wldgetag());
}

int
tshlist(                           /* tagspc handler for file lists        */
int tshcod)
{
     int rc=0;
     char *stg,fname[FLFILENM];

     fluoff(usrnum);
     setmbk(flmsg);
     stzcpy(fname,ftgptr->tagspc,FLFILENM);
     strupr(fname);
     switch (tshcod) {
     case TSHDSC:                      /* Describe tagspec in English */
          if ((stg=strchr(fname,'.')) != NULL) {
               *stg='\0';
          }
          if (sameas(fname,"FILES!")) {
               strcpy(tshmsg,"list of files in all Libraries");
          }
          else {
               sprintf(tshmsg,"list of files in the %s Library",fname);
          }
          break;
     case TSHVIS:                                  /* Visible to this user? */
          *tshmsg='\0';
          rc=1;
          break;
     case TSHBEG:              /* Begin download, check permission, reserve */
          sprintf(tshmsg,"%s\\%s",copydir,fname);
          if ((stg=strchr(fname,'.')) != NULL) {
               *stg='\0';
          }
          sprintf(ftfscb->fname,"%s.LST",fname);
          rc=1;
          break;
     case TSHEND:             /* End complete download of a file, unreserve */
          sv.dwnlds++;
          shocst("LIBRARY FILE DOWNLOAD","User %s download %s\\%s",
                           usaptr->userid,copydir,fname);
          unlink(spr("%s\\%s",copydir,fname));
          break;
     case TSHFIN:                           /* Finish file transfer session */
          retoin();
          rc=1;
          break;
     case TSHSKP:                     /* Skip incomplete download of a file */
          shocst("LIBRARY FILE DOWNLOAD ABORTED",
             "User %s aborted dnld of %s\\%s",usaptr->userid,copydir,fname);
     }
     return(rc);
}

int
tshlib(                            /* tagspc handler for File Libraries      */
int tshcod)
{
     static char tmsgtmp[200];
     struct fllib *libptr=NULL;
     FILE *fp=NULL;
     int rc=0;
     char *stg;
     long chg;

     fluoff(usrnum);
     setmbk(flmsg);
     fltagf=1;
     switch (tshcod) {
     case TSHSCN:
          if ((rc=dlbgnwld()) != 0) {
               movmem(ftgptr->tagspc,tshmsg,TSLENG);
          }
          break;
     case TSHNXT:
          if ((rc=wldgetag()) != 0) {
               movmem(ftgptr->tagspc,tshmsg,TSLENG);
          }
          else {
               spc->status=-spc->status;
          }
          break;
     case TSHDSC:                      /* Describe tagspec in English */
          setmem(tmsgtmp,200,0);
          if (spc->filname[0] == '*') {
               sprintf(tmsgtmp,"%d file%s matching \"%s\" in the %s Library",
                  spc->nwld,(spc->nwld > 1 ? "s" : ""),spc->wldspc,
                  spc->reflib);
          }
          else {
               sprintf(tmsgtmp,"file %s in the %s Library",spc->filname,
                  spc->reflib);
          }
          switch (spc->status) {
          case TGCOPY:
               if (spc->size > 0L) {
                    strcat(tmsgtmp,spr(" (Not Ready - %ld%%%% Copied)",
                       100L*spc->copied/spc->size));
               }
               else {
                    strcat(tmsgtmp," (Not Ready - Copying)");
               }
               break;
          case TGLOCAL:
          case TGDLTMP:
               stg=skpwht(dnlmin(spc->size));
               strcat(tmsgtmp,spr(" (%sK, %s minute",fzer(spc->size/1024L),
                  stg));
               if (atol(stg) > 1L) {
                    strcat(tmsgtmp,"s");
               }
               chg=dnlchg(spc->size,spc->libname);
               if (chg > 0L) {
                    strcat(tmsgtmp,spr(", %ld credits",chg));
               }
               strcat(tmsgtmp,")");
               break;
          case TGINWAIT:
               strcat(tmsgtmp," (Not Ready - Queued)");
               break;
          case TGOTHER:
          case TGSERVER:
               strcat(tmsgtmp," (Not Ready - Retrieving)");
               break;
          case -TGLOCAL:
          case -TGDLTMP:
               strcat(tmsgtmp," (Removing)");
               break;
          default:
               if (spc->status < TGEMPTY) {
                    strcat(tmsgtmp," (Aborted - Removing)");
               }
          }
          stzcpy(tshmsg,tmsgtmp,TSHLEN);
          break;
     case TSHVIS:                                  /* Visible to this user? */
          *tshmsg='\0';
          rc=1;
          if (spc->filname[0] == '*') {
               break;
          }
          if (spc->libname[0] != '\0') {
               libptr=libfind(spc->libname);
          }
          if (libptr != NULL
             && (spc->status == TGLOCAL || spc->status == TGDLTMP)) {
               if (libptr->flags&FLGCBD) {
                    sprintf(tshmsg,"%s\\%s\\%s",copydir,libptr->libname,
                       spc->filname);
               }
               else {
                    sprintf(tshmsg,"%s\\%s",libpath(libptr),spc->filname);
               }
               if ((fp=fopen(tshmsg,FOPRB)) != NULL) {
                    fread(tshmsg,1,TSHLEN,fp);
                    fclose(fp);
               }
               else {
                    if (spc->status == TGLOCAL) {
                         shocst("FILE LIBRARIES: MISSING FILE",
                                "%s\\%s is missing!",
                                libptr->libname,spc->filname);
                         prfmsg(MISFILEB,spc->filname);
                    }
                    else {
                         prfmsg(MISFILE,spc->filname);
                    }
                    outprf(usrnum);
                    clrprf();
                    rc=0;
               }
          }
          break;
     case TSHBEG:              /* Begin download, check permission, reserve */
          if (spc->libname[0] != '\0') {
               libptr=libfind(spc->libname);
               strcpy(tshmsg,tagtmpp);
               strcpy(ftfscb->fname,tagtmpf);
          }
          if (libptr == NULL) {
               strcpy(tshmsg,"file is not available for download");
               break;
          }
          switch (spc->status) {
          case TGINWAIT:
          case TGCOPY:
          case TGOTHER:
          case TGSERVER:
               strcpy(tshmsg,"file is not yet available for download");
               rc=-1;
               break;
          case TGLOCAL:
          case TGDLTMP:
               if (!sameas(spc->filname,"*")) {
                    if (libptr->flags&FLGCBD) {
                         sprintf(tshmsg,"%s\\%s\\%s",copydir,libptr->libname,
                            spc->filname);
                    }
                    else {
                         sprintf(tshmsg,"%s\\%s",libpath(libptr),spc->filname);
                    }
                    strcpy(ftfscb->fname,spc->filname);
               }
               if (!sameas(ftfpsp->code,"V")) {
                    if (!zapcrd((libptr->flags&FLGFDL) != 0L,spc->size,
                       libptr)) {
                         sprintf(tshmsg,
                           "You don't have enough credits to download %s",
                           ftfscb->fname);
                    }
                    else {
                         dnlcount(libptr->libname,ftfscb->fname,-1L);
                         rc=1;
                    }
               }
               else {
                    rc=1;
               }
          }
          break;
     case TSHEND:             /* End complete download of a file, unreserve */
          if (!sameas(ftfpsp->code,"V")) {
               sv.dwnlds++;
               if (spc->libname[0] != '\0') {
                    libptr=libfind(spc->libname);
               }
               if (libptr != NULL) {
                    dncfin(1,ftfscb->actbyt,libptr);
                    if (auditall || (libptr->flags&FLGADL)) {
                         shocst("LIBRARY FILE DOWNLOAD",
                            "User %s download %s\\%s",usaptr->userid,
                            spc->libname,ftfscb->fname);
                    }
               }
               rstcrd();
          }
          break;
     case TSHFIN:                           /* Finish file transfer session */
          retoin();
          rc=1;
          break;
     case TSHSKP:                     /* Skip incomplete download of a file */
          if (!sameas(ftfpsp->code,"V")) {
               if (spc->libname[0] != '\0') {
                    libptr=libfind(spc->libname);
               }
               if (libptr != NULL) {
                    dncfin(0,ftfscb->actbyt,libptr);
                    if (auditall || (libptr->flags&FLGADL)) {
                         shocst("LIBRARY FILE DOWNLOAD ABORTED",
                            "User %s aborted dnld of %s\\%s",
                            usaptr->userid,spc->libname,spc->filname);
                    }
               }
               rstcrd();
          }
          break;
     }
     return(rc);
}

long
dnlprc(                            /* price of download, one file + size   */
long siz,
struct fllib *libptr)
{
     return(haskey(nochgkey) || (usrptr->cltptr->flags&CRDXMT) ? 0L
      : libptr->dlchge+libptr->kdlchge*(siz/1024L));
}

void
fluoff(                            /* sets pointer to passed usrnum        */
int unum)                          /* usrnum to set it to                  */
{
     int saveu;

     if (unum < 0 || unum >= nterms) {
          unum=0;
     }
     saveu=usrnum;
     curfusr=usrnum=unum;
     curusr(usrnum);
     flo=(struct fluson *)ptrblok(flomem,usrnum);
     flu=&fluarr[usrnum];
     setlcsu(usrnum);
     curlib=setuaxs();
     setrsv(librsv);
     if (saveu != usrnum) {
          usrnum=saveu;
          if (usrnum >= 0 && usrnum < nterms) {
               curusr(usrnum);
          }
     }
}

void
srcoff(void)                       /* set up search pointer for cur fl user*/
{
     ASSERT(curfusr >= 0 && curfusr < nterms);
     srcu=areaptr(srcpool,srcharr[curfusr]);
     srcu->tn=(union srch *)srcu->unions;
     srcu->tp=(union srch *)(srcu->unions+maxjoin*sizeof(union srch));
     srcu->fn=(union srch *)(srcu->unions+maxjoin*sizeof(union srch)*2);
     srcu->fp=(union srch *)(srcu->unions+maxjoin*sizeof(union srch)*3);
}

void
scanlibs(void)                     /* set up Libraries/top-level Libraries */
{
     int loop=0;
     struct fllib *libptr;

     ncatlibs=0;
     while (loop < numoflib) {
          libptr=liboff(loop++);
          if (libptr->libname[0] != '\0') {
               if (scanlib(0,libptr)) {
                    ncatlibs++;
               }
          }
     }
}

struct fllib *
libfind(                           /* return pointer to Library            */
char *libname)                          /* ... Library name                */
{
     int loop;
     struct fllib *retval;

     for (loop=0 ; loop < numoflib ; loop++) {
          if (stricmp(libname,(retval=liboff(loop))->libname) == 0) {
               return(retval);
          }
     }
     return(NULL);
}

struct fllib *
liboff(                            /* return pointer to Library            */
int libnum)                        /* Library number                       */
{
     if (libnum >= numlib) {
          if (libmem2 == NULL) {
               libmem2=alcblok(numladd,sizeof(struct fllib));
          }
          return((struct fllib *)ptrblok(libmem2,libnum-numlib));
     }
     return((struct fllib *)ptrblok(libmem,libnum));
}

struct fllib *
setuaxs(void)                      /* set user access by key, returns ptr  */
{                                  /* to user's current Library or NULL    */
     int loop;
     struct fllib *retval,*flibp;

     if (flu->lib[0] == '\0'
        || (retval=libfind(flu->lib)) == NULL
        || !haslibkey(retval,retval->keyreq)) {
          setmem(flu->lib,FLNAMESZ,0);
          if ((retval=libfind(deflname)) != NULL && visilib(retval)) {
               stzcpy(flu->lib,deflname,FLNAMESZ);
          }
          else {
               for (loop=0 ; loop < numoflib ; loop++) {
                    if (valname((retval=liboff(loop))->libname)
                       && visilib(retval)) {
                         stzcpy(flu->lib,retval->libname,FLNAMESZ);
                         break;
                    }
               }
          }
          flo->nlibaxs=-1;
     }
     if (flo->nlibaxs == -1) {
          flo->nlibaxs=0;
          for (loop=0 ; loop < numoflib ; loop++) {
               flibp=liboff(loop);
               if (haslibkey(flibp,flibp->keyreq)) {
                    flo->nlibaxs++;
               }
          }
     }
     return(flu->lib[0] != '\0' ? retval : NULL);
}

int
makwdlst(                          /* make keyword list                    */
char *stg,
char *filname)
{
     char filnmem[FLFILENM],*ext;

     descedit[0]='\0';
     stzcpy(filnmem,filname,FLFILENM);
     ext=strchr(strupr(filname=filnmem),'.');
     if (ext != NULL) {
          *ext++='\0';
          if (strlen(ext) == 3) {
               strcat(descedit,ext);
               strcat(descedit," ");
          }
     }
     if (strlen(filname) > 2) {
          strcat(descedit,filname);
          strcat(descedit," ");
     }
     stzcat(descedit,stg,DESCSIZ);
     strcrep(descedit,'\r',' ');
     strupr(descedit);
     darsdesc(MAXKWDS,' ');
     bubkwds();
     return(dargc > 0);
}

int
submit(                            /* submit a file for tagging, 1=success */
char *libname,                     /* works on current user                */
char *filname,
char *reflib,                           /* reference lib in case true lib  */
int view,                                    /* is invisible to user,      */
int quiet)                                   /* "" acceptable, best guess  */
{
     static struct fndblk fb;
     int loop,isdos=0,wild;
     char *proprlib,*pro;
     struct fllib *libptr;
     int who;

     outprf(usrnum);
     prf("");
     for (loop=0 ; loop < nterms ; loop++) {
          if (loop != usrnum) {
               flo=(struct fluson *)ptrblok(flomem,loop);
               if ((flo->flags&REUPLOD) && sameas(flo->miscfil,filname)) {
                    flu=&fluarr[loop];
                    if (sameas(flu->lib,libname)) {
                         if (!quiet) {
                              prfmsg(CANTAGF,filname,libname);
                         }
                         return(0);
                    }
               }
          }
     }
     fluoff(usrnum);
     if (!view && alrtag(libname,filname) > 0) {
          return(0);
     }
     libptr=libfind(libname);
     if ((who=w2read(libname,filname)) != NOCONFLICT) {
          if (!quiet) {
               if (isflop(libptr)) {
                    prfmsg(CANTAGR,filname,libname,uacoff(who)->userid);
               }
               else {
                    prfmsg(CANTAGF,filname,libname);
               }
          }
          return(0);
     }
     ftag=NULL;
     for (loop=0 ; loop < systags ; loop++) {
          if (tagoff(loop)->status == TGEMPTY) {
               ftag=tagoff(loop);
               break;
          }
     }
     if (ftag == NULL) {
          if (!quiet) {
               prfmsg(CANTAGN,filname,libname);
          }
          return(0);
     }
     setbtv(flfdat);
     if (libptr != NULL && libptr->flags&FLGDOS) {
          isdos=1;
          if (*filname != '*'
             && !fnd1st(&fb,spr("%s\\%s",libpath(libptr),filname),0)) {
               if (!quiet) {
                    prfmsg(CANTAGF,filname,libname);
               }
               rstbtv();
               return(0);
          }
     }
     else if (*filname != '*'
        && !acqbtv(NULL,compkey(libname,filname),COMPLF)) {
          if (!quiet) {
               prfmsg(CANTAGF,filname,libname);
          }
          rstbtv();
          return(0);
     }
     rstbtv();
     proprlib=(libptr != NULL && !isdos ? visxlib(libptr,reflib) : libname);
     ftag->usrnum=usrnum;
     stzcpy(ftag->libname,libname,FLNAMESZ);
     stzcpy(ftag->reflib,proprlib,FLNAMESZ);
     stzcpy(ftag->filname,filname,FLFILENM);
     if (*filname == '*') {
          ftag->nwld=flo->wtnum;
          ftag->size=flo->wtsiz;
          stzcpy(ftag->wldspc,flo->miscfil,FLFILENM);
     }
     else {
          ftag->size=isdos ? fb.size : flf->siz;
          stzcpy(ftag->udate,isdos ? dat2srt(ncdate(today())) :
             flf->udate,DATESZ);
     }
     ftag->tagtime=time(NULL);
     if (ftag->libname[0] == '\0' || (libptr=libfind(ftag->libname)) == NULL) {
          ftag->status=-1;
          if (quiet) {
               clrprf();
          }
          return(0);
     }
     ftag->status=((libptr->flags&FLGCBD)
        ? (libptr->path[0] != '\0' ? TGINWAIT : TGOTHER) : TGLOCAL);
     if (ftgnew()) {
          spc=ftag;
          wild=sameas(ftag->filname,"*");
          ftgptr->flags=(wild ? FTGWLD : 0);
          ftgptr->flags|=(view != NOTVIEW && !wild ? 0 : FTGABL);
          ftgptr->tshndl=tshlib;
          if (view == NOTVIEW) {
               ftgsbm("TQ");
          }
          else {
               pro="";
               if (flu->askme && view == DOWNLOD && flo->savprot[0] == '\0') {
                    ftgptr->flags|=FTGABL;
                    ftgsbm("TQ");
                    setmbk(flmsg);
                    prfmsg(CLRSCN);
                    prfmsg(DEFRIPM);
                    switch (usrptr->substt) {
                    case FILELST0:
                         prfmsg(usrptr->substt=LOGOFF3);
                         break;
                    case ENTDLFIL:
                         usrptr->substt=LOGOFF;
                         break;
                    default:
                         prfmsg(usrptr->substt=LOGOFF2);
                    }
                    outprf(usrnum);
                    prf("");
                    rstmbk();
                    echonu(usrnum);
                    btutrg(usrnum,0);
                    flo->pflg=ftuptr->numftg;
                    return(1);
               }
               if (pro[0] == '\0') {
                    pro=protocol();
               }
               if (flo->savprot[0] != '\0') {
                    pro=spr("%s",flo->savprot);
                    flo->savprot[0]='\0';
               }
               if (view == CMPVIEW) {
                    pro="V";
                    prf("[0;1;36m");
               }
               else if (view == LSTVIEW) {
                    pro="L";
               }
               else if (view == ANSVIEW) {
                    pro="A";
               }
               else if (view == CSTVIEW) {
                    prf("");
                    (*flo->rouptr)(ftag);
                    outprf(usrnum);
                    prf("");
                    return(1);
               }
               if (sameto("T",pro) && !wild) {
                    pro="";
               }
               if (pro[0] == '\0') {
                    pro="?";
               }
               ftgsbm(pro);
               if (usrptr->state != flstt) {
                    prf("");
                    outprf(usrnum);
               }
               else {
                    prf("");
                    if (!sameto("T",pro)) {
                         return(0);
                    }
               }
          }
     }
     if (quiet) {
          clrprf();
     }
     return(1);
}

int
updusr(                            /* update user field to disk            */
struct fluser *fluptr)
{
     int retval=-1;

     stzcpy(flusmem.userid,usaptr->userid,UIDSIZ);
     stzcpy(flusmem.modnam,"File Library",MNMSIZ);
     movmem(fluptr,&flusmem.udata,sizeof(struct fluser));
     setbtv(genbb);
     if (acqbtv(NULL,&flusmem,0)) {
          upvbtv(&flusmem,sizeof(struct flusave));
          retval=1;
     }
     else if (flusmem.userid[0] != '\0') {
          invbtv(&flusmem,sizeof(struct flusave));
          retval=0;
     }
     rstbtv();
     return(retval);
}

char *
userdir(                           /* return temp directory or filename    */
int which)                         /* 0=dir 1=user ul 2=sdi 3=diz          */
{
     switch (which) {
     case 1:
          return(spr("%s\\UPLOAD%02x",tempdir,channel[usrnum]));
     case 2:
          return(spr("%s\\DESC%02x.SDI",tempdir,channel[usrnum]));
     case 3:
          return(spr("%s\\DESC%02x.DIZ",tempdir,channel[usrnum]));
     }
     return(spr("%s",tempdir));
}

int
valname(                           /* is passed name valid for a Library?  */
char *libname)
{
     int len;
     char *cptr;

     len=strlen(cptr=libname);
     if (len < 1 || len > 8) {
          return(0);
     }
     while (*cptr != '\0') {
          if (!isfiln(*cptr,".")) {
               return(0);
          }
          cptr++;
     }
     if (rsvnam(libname) || profane(libname)) {
          return(0);
     }
     return(1);
}

int
fileditv(                          /* verify FSD edit file session         */
int fldno,
char *answer)
{
     char nts[9];
     long ndl;

     fluoff(usrnum);
     switch (fldno) {
     case FFDULT:
          stzcpy(nts,answer,9);
          if (nts[1] == ':') {
               nts[0]='0';
               stzcpy(&nts[1],answer,8);
          }
          if (strlen(nts) == 5) {
               strcat(nts,":00");
          }
          stzcpy(answer,nctime(dctime(nts)),9);
          if (!sameas(nts,answer)) {
               setmbk(flmsg);
               stzcpy(fsdemg,getmsg(BADTIM),MAXHLP);
               rstmbk();
               return(VFYREJ);
          }
          break;
     case FFDDL:
          ndl=atol(answer);
          if (ndl < 0L) {
               ndl=0;
          }
          if (ndl > 9999999L) {
               ndl=9999999L;
          }
          strcpy(answer,l2as(ndl));
          break;
     case FFDULD:
          fsdemg[0]='\0';
          setmbk(flmsg);
          if (atoi(answer) == 0) {
               stzcpy(fsdemg,getmsg(BADDAT),MAXHLP);
          }
          else {
               if (answer[1] == '/') {
                    strcpy(answer,spr("0%s",answer));
               }
               if (answer[4] == '/') {
                    strcpy(&answer[3],spr("0%s",&answer[3]));
               }
               if (!sameas(answer,srt2dat(dat2srt(answer)))) {
                    stzcpy(fsdemg,getmsg(BADDAT),MAXHLP);
               }
               else if (strcmp(dat2srt(answer),dat2srt(ncdate(today()))) < 0) {
                    stzcpy(fsdemg,getmsg(BADDAT2),MAXHLP);
               }
          }
          rstmbk();
          if (fsdemg[0] != '\0') {
               return(VFYREJ);
          }
          break;
     case FFDUSER:
          if (sameas(answer,"Sysop")) {
               stzcpy(answer,"Sysop",UIDSIZ);
               break;
          }
          setbtv(accbb);
          if (!acqbtv(&acctmp,answer,0)) {
               setmbk(flmsg);
               stzcpy(fsdemg,getmsg(BADUID),MAXHLP);
               rstmbk();
               rstbtv();
               return(VFYREJ);
          }
          else {
               stzcpy(answer,acctmp.userid,UIDSIZ);
          }
          rstbtv();
          break;
     case FFDTXT00:
     case FFDTXT01:
     case FFDTXT02:
     case FFDTXT03:
     case FFDTXT04:
     case FFDTXT05:
     case FFDTXT06:
     case FFDTXT07:
     case FFDTXT08:
     case FFDTXT09:
     case FFDTXT10:
     case FFDTXT11:
     case FFDTXT12:
     case FFDTXT13:
     case FFDTXT14:
          if (answer[0] == '\0' && (fsdscb->newans+fldptr->ansoff)[0]== '\0') {
               fldno=fsdscb->crsfld=FFDEXIT;
          }
          else if (profane(answer)) {
               setmbk(flmsg);
               stzcpy(fsdemg,getmsg(PROFDESC),MAXHLP);
               rstmbk();
               return(VFYREJ);
          }
     }
     return(vfyadn(fldno,answer));
}

int
lsinp(void)                        /* main input handler                   */
{
     fluoff(usrnum);
     setftu();
     setmbk(flmsg);
     if (curlib == NULL || flu->lib[0] == '\0') {
          updusr(flu);
          if (usrptr->substt == 0) {
               prfmsg(NOAXESS);
          }
          return(0);
     }
     if (usrptr->flags&ABOIP) {
          switch (usrptr->substt) {
          case VIEWASC:
               prfmsg(usrptr->substt=VIEWASC);
               break;
          case ENTRETT:
               buildview(flo->hold);
               break;
          case FILELST0:
               rebuild(2);
               break;
          case TAGMAN:
               retmenu(0);
               mainmenu();
               prfmsg(MMENU);
               break;
          case LIBLST:
               usrptr->substt=flo->retstt;
               break;
          default:
               clrprf();
          }
          prf("");
          outprf(usrnum);
          prf("");
     }
     if (!(usrptr->flags&INJOIP) && !(usrptr->flags&ABOIP)) {
          do {
               bgncnc();
               if ((flo->flags&ENTRCNC) && isripu()) {
                    prfmsg(DEFRIPM);
               }
               if (usrptr->substt == 0) {
                    cncchr();
                    usrptr->flags&=~X2MAIN;
                    usrptr->substt=MMENU;
                    flo->flags&=~ABOUPRF;
                    flo->flags|=LONGDSC;
                    flo->flags&=~INMVCPY;
                    if (morcnc() == '&') {
                         flo->flags|=ABOUPRF;
                         flo->flags|=ENTRCNC;
                         cncchr();
                    }
               }
               else {
                    switch (inpfunc()) {
                    case 0:
                         flu->cdayin=cofdat(today());
                         deresume();
                         return(0);
                    case 2:
                         return(1);
                    }
                    fluoff(usrnum);
                    setftu();
               }
               if (curlib == NULL) {
                    return(0);
               }
               if (nxtcmd == margv[0]) { /* if inp handler didn't use any, */
                    cncall();            /* don't let it go around again   */
               }
          } while (!endcnc());
          flo->flags&=~ENTRCNC;
     }
     promptin();
     return(1);
}

void
deresume(void)                     /* cancel possibility of resuming list  */
{
     flo->tagk='A'-1;
}

int
lslon(void)                        /* system logon routine                 */
{
     switch (usrptr->substt) {
          case -1:
               return(1);
          case -2:
               rmdir(userdir(1));
               return(0);
     }
     if (usrptr->flags&WSGCSU) {
          fluoff(usrnum);
     }
     else {
          genlon();
          if (curlib != NULL && (notiflop || notifsys)) {
               notify(1);
          }
     }
     if (ulfileq(1)) {
          setmbk(flmsg);
          prfmsg(HANGON);
          rstmbk();
          spinner();
          usrptr->substt=-1;
          btuinj(usrnum,CYCLE);
          outprf(usrnum);
          return(1);
     }
     rmdir(userdir(1));
     return(0);
}

void
genlon(void)                       /* C/S - A/A generic logon routine parts*/
{
     flu=&fluarr[usrnum];
     flo=(struct fluson *)ptrblok(flomem,usrnum);
     setmem(flu,sizeof(struct fluser),0);
     setmem(flo,sizeof(struct fluson),0);
     flo->nlibaxs=-1;
     fluoff(usrnum);
     setbtv(genbb);
     stzcpy(flusmem.userid,usaptr->userid,UIDSIZ);
     stzcpy(flusmem.modnam,"File Library",MNMSIZ);
     if (!acqbtv(&flusmem,&flusmem,0)) {
          stzcpy(flusmem.userid,usaptr->userid,UIDSIZ);
          stzcpy(flusmem.modnam,"File Library",MNMSIZ);
          setmem(&flusmem.udata,sizeof(struct fluser),0);
     }
     movmem(&flusmem.udata,flu,sizeof(struct fluser));
     curlib=setuaxs();
     rstbtv();
     updusr(flu);
     if (flu->newdate[0] == '\0') {
          stzcpy(flu->newdate,dat2srt(ncdate(datofc(cofdat(today())-3))),
             DATESZ);
     }
     stzcpy(flo->newdate,flu->newdate,DATESZ);
}

int
srcincx(void)                      /* is screen not full of search elements*/
{
     int loop;

     if (flo->flags&REVSRCH) {
          if (flo->atend != LSTFILS && flo->atend != LSTBACK) {
               return(0);
          }
          for (loop=NLISTER-1 ; loop >= 0 ; loop--) {
               if (flo->tdline[loop] == 0) {
                    return(loop+1);
               }
          }
          return(0);
     }
     if (flo->atend != LSTFILS && flo->atend != LSTFORE) {
          return(0);
     }
     for (loop=0 ; loop < NLISTER ; loop++) {
          if (flo->tdline[loop] == 0) {
               return(loop+1);
          }
     }
     return(0);
}

int
srcgood(void)                      /* display, reject, or abort search?    */
{
     int retval=0;
     struct fllib *libptr=NULL;
     char wildkey[FLFILENM];
     union srch *current,*new,*old;

     srcoff();
     if (flo->keymeth > 2) {
          retval=-1;
     }
     if (srcu->nlibs > 0) {
          new=(flo->flags&REVSRCH) ? &srcu->tp[srcindex]
             : &srcu->tn[srcindex];
          old=(flo->flags&REVSRCH) ? &srcu->fp[srcindex]
             : &srcu->fn[srcindex];
          movmem(new,old,sizeof(union srch));
          current=&srcactive;
          switch (flo->keymeth) {
          case 3:
               if (((flo->flags&REVSRCH)
                  && altbtv(NULL,&current->key0,COMPLUF)
                  || !(flo->flags&REVSRCH)
                  && agtbtv(NULL,&current->key0,COMPLUF))
                  && sameas(flf->libname,current->key0.libname)) {
                    stzcpy(new->key0.libname,flf->libname,FLNAMESZ);
                    stzcpy(new->key0.udate,flf->udate,DATESZ);
                    stzcpy(new->key0.filname,flf->filname,FLFILENM);
               }
               else {
                    new->key0.udate[0]=(flo->flags&REVSRCH) ? 0x00 : 0xFF;
               }
               if (!acqbtv(NULL,&srcactive.key0,COMPLUF)) {
                    return(0);
               }
               break;
          case 4:
          case 7:
          case 8:
               if (((flo->flags&REVSRCH)
                  && altbtv(NULL,&current->key1,COMPLF)
                  || !(flo->flags&REVSRCH)
                  && agtbtv(NULL,&current->key1,COMPLF))
                  && sameas(flf->libname,current->key1.libname)) {
                    stzcpy(new->key1.libname,flf->libname,FLNAMESZ);
                    stzcpy(new->key1.filname,flf->filname,FLFILENM);
               }
               else {
                    new->key1.filname[0]=(flo->flags&REVSRCH) ? 0x00 : 0xFF;
               }
               if (!acqbtv(NULL,&srcactive.key1,COMPLF)) {
                    return(0);
               }
               break;
          }
     }
     switch (flo->keymeth) {
     case 0:
          if (sameas(flf->libname,flo->u.key0.libname)) {
               retval=0;
               if ((libptr=libfind(flo->u.key0.libname)) != NULL) {
                    retval=(notapped(flf->udate) && !isflop(libptr) ? -1 : 1);
               }
          }
          if ((flo->flags&DATESRC) && stricmp(flf->udate,flo->newdate) > 0) {
               retval=0;
          }
          break;
     case 1:
          if (sameas(flf->libname,flo->u.key1.libname)) {
               retval=0;
               if ((libptr=libfind(flo->u.key1.libname)) != NULL) {
                    retval=(notapped(flf->udate) && !isflop(libptr) ? -1 : 1);
               }
          }
          break;
     case 2:
          libptr=libfind(kyd->libkey);
          if (libptr == NULL || kyd->libkey[0] == '\0'
             || !strlen(visxlib(libptr,flo->u.key2.libname))) {
               retval=-1;
          }
          else {
               retval=1;
          }
          if (!sameto(flo->u.key2.origkey,kyd->keyword)) {
               retval=0;
          }
          if (retval == 1) {
               setbtv(flfdat);
               if (!acqbtv(NULL,compkey(kyd->libkey,kyd->filname),COMPLF)) {
                    retval=-1;
               }
               if (notapped(flf->udate) && !isflop(libptr)) {
                    retval=-1;
               }
               if ((retval == 1) && (flo->kcount > 1)) {
                    retval=keysin(0);
               }
               rstbtv();
          }
          break;
     case 3:
          keylib(flf->libname,flo->u.key3.libname,&retval);
          if ((flo->flags&DATESRC) && stricmp(flf->udate,flo->newdate) > 0) {
               retval=0;
          }
          break;
     case 4:
          keylib(flf->libname,flo->u.key4.libname,&retval);
          if (retval == 1 && flo->styp == 6) {
               retval=keyfig();
          }
          break;
     case 5:
          keylib(flf->libname,flo->u.key5.libname,&retval);
          break;
     case 6:
          if (sameas(flf->libname,flo->u.key1.libname)) {
               if ((libptr=libfind(flo->u.key1.libname)) == NULL) {
                    retval=0;
               }
               else if (notapped(flf->udate) && !isflop(libptr)) {
                    retval=-1;
               }
               else if (!compwild(flo->miscfil,flf->filname)) {
                    retval=(strlen(wldstuff(wildkey))
                         && !sameto(wildkey,flf->filname) ? 0 : -1);
               }
               else {
                    retval=1;
               }
          }
          else {
               retval=0;
          }
          break;
     case 7:
          keylib(flf->libname,flo->u.key4.libname,&retval);
          if (retval == 1 && !compwild(flo->miscfil,flf->filname)) {
               retval=(strlen(wldstuff(wildkey))
                    && !sameto(wildkey,flf->filname) ? 0 : -1);
          }
          break;
     case 8:
          keylib(flf->libname,flo->u.key4.libname,&retval);
          if (retval == 1 && !sameas(flf->filname,flo->u.key4.exactfn)) {
               retval=0;
          }
          break;
     case 9:
          flo->atend=LSTFORE;
          retval=1;
     }
     return(retval);
}

int
srcnext(void)                      /* get next search element              */
{
     int flag1,flag2;

     srcoff();
     flag1=((flo->flags&REVSRCH) ? 1 : 0);
     flag2=((flo->flags&NXTEQUL) ? 1 : 0);
     flo->flags&=~NXTEQUL;
     setbtv(flfdat);
     switch (flo->keymeth) {
     case 0:
          return(nextrec(flag1,flag2,NULL,flo->u.key0.libname,COMPLUF));
     case 1:
     case 6:
          return(nextrec(flag1,flag2,NULL,flo->u.key1.libname,COMPLF));
     case 2:
          setbtv(flkdat);
          return(nextrec(flag1,flag2,NULL,flo->u.key2.keyword,COMPKFL));
     case 3:
          if (srcu->nlibs > 0) {
               return(nextsrc(flag2) != NULL);
          }
          return(nextrec(flag1,flag2,NULL,flo->u.key3.udate,COMPUF));
     case 4:
     case 7:
     case 8:
          if (srcu->nlibs > 0) {
               return(nextsrc(flag2) != NULL);
          }
          return(nextrec(flag1,flag2,NULL,flo->u.key4.filname,COMPFL));
     case 5:
          return(nextrec(!flag1,flag2,NULL,flo->u.key5.numdls,COMPNLF));
     case 9:
          if (flo->u.key6.filname[0] == '\0') {
               return(fnd1st(&flo->fb,flo->srcpath,0));
          }
          return(fndnxt(&flo->fb));
     }
     return(0);
}

union srch *
nextsrc(                           /* return pointer to next join record   */
int nxteq)
{
     union srch *best=NULL,*current;
     int loop,si;
     char badchr;

     srcoff();
     if (nxteq) {
          for (loop=0 ; loop < srcu->nlibs ; loop++) {
               current=(flo->flags&REVSRCH) ? &srcu->tp[loop]
                  : &srcu->tn[loop];
               switch (flo->keymeth) {
               case 3:
                    if (sameas(current->key0.libname,flo->u.key3.libkey)) {
                         if (sameas(current->key0.udate,flo->u.key3.udate)
                            && sameas(current->key0.filname,
                            flo->u.key3.filname)) {
                              srcindex=loop;
                              movmem(current,&srcactive,sizeof(union srch));
                              return(current);
                         }
                         loop=maxjoin;
                    }
                    break;
               case 4:
               case 7:
               case 8:
                    if (sameas(current->key1.libname,flo->u.key4.libkey)) {
                         if (sameas(current->key1.filname,
                            flo->u.key4.filname)) {
                              srcindex=loop;
                              movmem(current,&srcactive,sizeof(union srch));
                              return(current);
                         }
                         loop=maxjoin;
                    }
                    break;
               }
          }
     }
     badchr=(flo->flags&REVSRCH) ? 0x00 : 0xFF;
     for (loop=0 ; loop < srcu->nlibs ; loop++) {
          current=(flo->flags&REVSRCH) ? &srcu->tp[loop] : &srcu->tn[loop];
          switch (flo->keymeth) {
          case 3:
               if (current->key0.libname[0] != '\0') {
                    if (best != NULL) {
                         si=stricmp(best->key0.udate,current->key0.udate);
                         if (si == 0) {
                              si=stricmp(best->key0.filname,
                                 current->key0.filname);
                              if (si == 0) {
                                   si=stricmp(best->key0.libname,
                                      current->key0.libname);
                              }
                         }
                         if ((flo->flags&REVSRCH) && si < 0
                            || !(flo->flags&REVSRCH) && si > 0) {
                              if (current->key0.udate[0] != badchr) {
                                   best=current;
                              }
                         }
                    }
                    else if (current->key0.udate[0] != badchr) {
                         best=current;
                    }
               }
               break;
          case 4:
          case 7:
          case 8:
               if (current->key1.libname[0] != '\0') {
                    if (best != NULL) {
                         si=stricmp(best->key1.filname,current->key1.filname);
                         if (si == 0) {
                              si=stricmp(best->key1.libname,
                                 current->key1.libname);
                         }
                         if ((flo->flags&REVSRCH) && si < 0
                            || !(flo->flags&REVSRCH) && si > 0) {
                              if (current->key1.filname[0] != badchr) {
                                   best=current;
                              }
                         }
                    }
                    else if (current->key1.filname[0] != badchr) {
                         best=current;
                    }
               }
               break;
          }
          if (best == current) {
               srcindex=loop;
          }
     }
     if (best != NULL) {
          movmem(best,&srcactive,sizeof(union srch));
     }
     return(best);
}

void
resetkey(                          /* re-set key properly for direction.   */
char *libname,
char *filname)
{
     setbtv(flfdat);
     if (!acqbtv(NULL,compkey(libname,filname),COMPLF)) {
          setmem(flf,sizeof(struct flfile),0);
     }
     switch (flo->keymeth) {
     case 0:
          stzcpy(flo->u.key0.libname,flf->libname,FLNAMESZ);
          stzcpy(flo->u.key0.filname,flf->filname,FLFILENM);
          stzcpy(flo->u.key0.udate,flf->udate,DATESZ);
          break;
     case 1:
     case 6:
          stzcpy(flo->u.key1.libname,flf->libname,FLNAMESZ);
          stzcpy(flo->u.key1.filname,flf->filname,FLFILENM);
          break;
     case 2:
          setbtv(flkdat);
          stzcpy(flo->u.key2.libkey,flf->libname,FLNAMESZ);
          stzcpy(flo->u.key2.filname,flf->filname,FLFILENM);
          break;
     case 3:
          stzcpy(flo->u.key3.filname,flf->filname,FLFILENM);
          stzcpy(flo->u.key3.udate,flf->udate,DATESZ);
          stzcpy(flo->u.key3.libkey,flf->libname,FLNAMESZ);
          break;
     case 4:
     case 7:
     case 8:
          stzcpy(flo->u.key4.libkey,flf->libname,FLNAMESZ);
          stzcpy(flo->u.key4.filname,flf->filname,FLFILENM);
          break;
     case 5:
          sprintf(flo->u.key5.numdls,NUMSZP,atol(flf->numdls));
          stzcpy(flo->u.key5.libkey,flf->libname,FLNAMESZ);
          stzcpy(flo->u.key5.filname,flf->filname,FLFILENM);
     }
}

void
rset(void)                         /* find key to reset to and do it       */
{
     int loop;

     if (!(flo->flags&REVSRCH)) {
          for (loop=NLISTER-1 ; loop >= 0 ; loop--) {
               if (flo->tdline[loop] != 0) {
                    resetkey(flo->tags[loop].libname,flo->tags[loop].filname);
                    break;
               }
          }
     }
     else {
          for (loop=0 ; loop < NLISTER ; loop++) {
               if (flo->tdline[loop] != 0) {
                    resetkey(flo->tags[loop].libname,flo->tags[loop].filname);
                    break;
               }
          }
     }
}

int
dosinsrt(                          /* insert file into DOS-Only Library    */
struct fllib *libptr,              /* Library                              */
char *path,                        /* path where file is stored            */
struct fndblk *f)                  /* stuff including filename info        */
{
     stzcpy(flo->srcpath,spr("%s\\%s",fixpath(path),f->name),PATHSIZE);
     stzcpy(flo->destpath,spr("%s\\%s",libpath(libptr),f->name),PATHSIZE);
     if (sameas(flo->srcpath,flo->destpath)) {
          return(1);
     }
     unlink(flo->destpath);
     if (rename(flo->srcpath,flo->destpath) != 0) {
          flo->fsrc=fopen(flo->srcpath,FOPRB);
          flo->fdst=fopen(flo->destpath,FOPWB);
          if (flo->fsrc == NULL) {
               if (flo->fdst != NULL) {
                    fclose(flo->fdst);
                    flo->fdst=NULL;
               }
               return(-2);
          }
          if (flo->fdst == NULL) {
               if (flo->fsrc != NULL) {
                    fclose(flo->fsrc);
                    flo->fsrc=NULL;
               }
               return(-2);
          }
          return(2);
     }
     return(1);
}

int
insfile(                           /* insert file into file database       */
struct fllib *libptr,              /* Library                              */
char *path,                        /* path where file is stored            */
struct fndblk *f,                  /* stuff including filename info        */
char *appdate,                     /* approval date, NOTAPPED if unappved  */
char *u,                           /* uploaded by                          */
char *desc)                        /* var length description               */
{
     struct taglib *tag;
     int retval=1,loop;
     char *key;

     if (libptr->flags&FLGDOS) {
          return(dosinsrt(libptr,path,f));
     }
     setbtv(flfdat);
     if (unifile && agebtv(NULL,key=rcompkey("",f->name),COMPFL)
                 && sameas(flf->filname,key)) {
          rstbtv();
          return(-1);
     }
     if (acqbtv(NULL,key=compkey(libptr->libname,f->name),COMPLF)) {
          if (!(flo->flags&REUPLOD)) {
               flo->fsrc=fopen(spr("%s\\%s",libpath(libptr),f->name),FOPRB);
               if (flo->fsrc != NULL) {
                    fclose(flo->fsrc);
                    flo->fsrc=NULL;
                    rstbtv();
                    return(-1);
               }
               else if (acqbtv(NULL,key,COMPLF)) {
                    delbtv();
               }
          }
     }
     else {
          if (flo->flags&REUPLOD) {
               flo->flags&=~REUPLOD;
          }
          unlink(spr("%s\\%s",libpath(libptr),f->name));
     }
     if (!(flo->flags&REUPLOD)) {
          setmem(flf,FLFILREC+DESCSIZ,0);
          stzcpy(flf->libname,libptr->libname,FLNAMESZ);
          stzcpy(flf->filname,f->name,FLFILENM);
          stzcpy(flf->udate,appdate,DATESZ);
          flf->utime=now();
          flf->siz=f->size;
          stzcpy(flf->fdate,dat2srt(ncdate(f->date)),DATESZ);
          flf->tim=f->time;
          stzcpy(flf->numdls,spr(NUMSZP,0L),NUMSZ);
          stzcpy(flf->ulby,u,UIDSIZ);
          stzcpy(flf->desc,desc,DESCSIZ);
          invbtv(NULL,FLFILREC+strlen(desc)+1);
     }
     else {
          for (loop=0 ; loop < systags ; loop++) {
               tag=tagoff(loop);
               if (tag->status > TGEMPTY && tag->usrnum != usrnum
                  && sameas(tag->filname,f->name)
                  && sameas(tag->libname,libptr->libname)) {
                    prf("\rAnother user "
                        "has this file, %s, tagged for download!\r",f->name);
                    outprf(usrnum);
                    prf("");
                    stzcpy(flf->filname,f->name,FLFILENM);
                    stzcpy(flf->libname,libptr->libname,FLNAMESZ);
                    return(-1);
               }
          }
          libptr->numfiles--;
          libptr->totbytes-=clfit(flf->siz,libptr->cluster);
          if (notapped(flf->udate)) {
               libptr->appwait--;
          }
          chuldate(libptr,flf->udate,0);
          stzcpy(flf->udate,appdate,DATESZ);
          flf->utime=now();
          flf->siz=f->size;
          stzcpy(flf->fdate,dat2srt(ncdate(f->date)),DATESZ);
          flf->tim=f->time;
          stzcpy(flf->ulby,u,UIDSIZ);
          upvbtv(NULL,FLFILREC+strlen(flf->desc)+1);
     }
     stzcpy(flo->key1.libname,libptr->libname,FLNAMESZ);
     stzcpy(flo->key1.filname,f->name,FLFILENM);
     stzcpy(flo->srcpath,spr("%s\\%s",fixpath(path),f->name),PATHSIZE);
     stzcpy(flo->destpath,spr("%s\\%s",libpath(libptr),f->name),PATHSIZE);
     if (!sameas(flo->srcpath,flo->destpath)) {
          unlink(flo->destpath);
          if (rename(flo->srcpath,flo->destpath) != 0) {
               retval=2;
               flo->fsrc=fopen(flo->srcpath,FOPRB);
               flo->fdst=fopen(flo->destpath,FOPWB);
               if (flo->fsrc == NULL) {
                    retval=-2;
               }
               if (flo->fdst == NULL) {
                    retval=-2;
                    if (flo->fsrc != NULL) {
                         fclose(flo->fsrc);
                         flo->fsrc=NULL;
                    }
               }
               else if (retval == -2) {
                    fclose(flo->fdst);
                    flo->fdst=NULL;
               }
          }
          else {
               retval=1;
          }
     }
     else {
          retval=1;
     }
     libptr->numfiles++;
     libptr->totbytes+=clfit(f->size,libptr->cluster);
     if (notapped(appdate)) {
          libptr->appwait++;
     }
     chuldate(libptr,appdate,1);
     libptr->flags|=LIBCHN;
     return(retval);
}

int
keywordy(void)                     /* state handler                        */
{
     char *curkwd,*lastkwd;
     int loop;

     spinner();
     vdaptr=vdaoff(usrnum);
     if (flo->miscflag == -1) {
          setbtv(flkdat);
          if (!delkw("",flo->kwdy.filname,flo->kwdy.libkey)) {
               flo->miscflag=0;
               setbtv(flfdat);
               if (acqbtv(NULL,compkey(flo->kwdy.libkey,flo->kwdy.filname),
                  COMPLF) && makwdlst(flf->desc,flf->filname)) {
                    for (loop=0 ; loop < dargc ; loop++) {
                         stzcpy(&vdaptr[FLKEYSIZ*loop],dargv[loop],FLKEYSIZ);
                    }
                    flo->kcount=dargc;
               }
               rstbtv();
          }
     }
     else {
          if (flo->miscflag >= flo->kcount) {
               return(0);
          }
          lastkwd=(flo->miscflag ? &vdaptr[FLKEYSIZ*(flo->miscflag-1)] : "");
          curkwd=&vdaptr[FLKEYSIZ*flo->miscflag++];
          stzcpy(flo->kwdy.keyword,curkwd,FLKEYSIZ);
          if (!sameas(curkwd,lastkwd) && !samein(spr(" %s ",curkwd),nonkwds)) {
               setbtv(flkdat);
               insbtv(&flo->kwdy);
               rstbtv();
          }
     }
     prf("");
     outprf(usrnum);
     prf("");
     return(1);
}

int
libeditv(                          /* verify FSD edit Library session      */
int fldno,
char *answer)
{
     long lrange;
     char *stg;

     switch (fldno) {
     case FLDCBD:
          if (sameas(answer,"YES")) {
               fsdfxt(FLDPATH,stg=spr(""),PATHSIZE);
               if (stg[0] == '\0') {
                    setmbk(flmsg);
                    stzcpy(fsdemg,getmsg(SVRNOT),MAXHLP);
                    rstmbk();
               }
          }
          break;
     case FLDPATH:
          if (answer[0] != '\0' && !valpath(answer)) {
               setmbk(flmsg);
               stzcpy(fsdemg,getmsg(BADDIR),MAXHLP);
               rstmbk();
               return(VFYREJ);
          }
          strupr(checkdir(answer));
          break;
     case FLDPLOP:
          if (answer[0] == '\0') {
               break;
          }
          if (sameas(answer,"Sysop")) {
               stzcpy(answer,"Sysop",UIDSIZ);
               break;
          }
          setbtv(accbb);
          if (!acqbtv(&acctmp,answer,0)) {
               setmbk(flmsg);
               stzcpy(fsdemg,getmsg(BADUID),MAXHLP);
               rstmbk();
               rstbtv();
               return(VFYREJ);
          }
          else {
               stzcpy(answer,acctmp.userid,UIDSIZ);
          }
          rstbtv();
          break;
     case FLDVIS:
     case FLDDLD:
     case FLDULD:
     case FLDOVW:
     case FLDAPU:
     case FLDLOP:
          if (answer[0] != '\0' && !keynam(strupr(answer))) {
               setmbk(flmsg);
               stzcpy(fsdemg,getmsg(BADKEY),MAXHLP);
               rstmbk();
               return(VFYREJ);
          }
          break;
     case FLDROY:
          if ((lrange=atol(answer)) < 0L || lrange > 100L) {
               setmbk(flmsg);
               stzcpy(fsdemg,getmsg(BADROY),MAXHLP);
               rstmbk();
               return(VFYREJ);
          }
          strcpy(answer,l2as(lrange));
          break;
     case FLDMFL:
     case FLDMBT:
     case FLDMUP:
          if ((lrange=atol(answer)) < 0L || lrange > 999999999L) {
               setmbk(flmsg);
               stzcpy(fsdemg,getmsg(BADLONG),MAXHLP);
               rstmbk();
               return(VFYREJ);
          }
     case FLDCDF:
     case FLDCDK:
          strcpy(answer,l2as(atol(answer)));
     }
     return(vfyadn(fldno,answer));
}

int
valpath(                           /* determine validity of a passed path  */
char *path)
{
     static struct fndblk fb;
     int flag;

     path=fixpath(path);
     flag=mkdir(path);
     if (flag != 0 && !fnd1st(&fb,path,FAMDIR)) {
          if (!samend(path,":\\.") && !samend(path,":/.")) {
               if (strlen(path) != 3 || !samend(path,":.")) {
                    return(0);
               }
          }
     }
     if (flag == 0) {
          rmdir(path);
     }
     return(1);
}

int
taglist(void)                      /* display taglist if there is one      */
{
     int loop,lastidx=-1,nlfiles=0,idx=-1,numdisp=0,lim,warntime=0,blimit,
         stoprf=0;
     long totsize=0L,totchg=0L,totsecs=0L,pbuf;
     struct fllib *lib;

     if (ftuptr->numftg > 0) {
          moploop();
          setftu();
          prfmsg(TAGLST);
          blimit=btuoba(usrnum)-1200;
          for (loop=0 ; loop < ftuptr->numftg ; loop++) {
               ftgptr=&ftgusr[loop];
               fltagf=0;
               ftgptr->tshndl(TSHDSC);
               setmbk(flmsg);
               if (fltagf) {
                    if (spc->status == TGLOCAL || spc->status == TGDLTMP) {
                         if ((lib=libfind(spc->libname)) != NULL) {
                              totsize+=spc->size;
                              totchg+=dnlchg(spc->size,spc->libname);
                              totsecs+=dnlsecl(spc->size,lib);
                              nlfiles++;
                         }
                         else {
                              spc->status=-spc->status;
                         }
                    }
                    if (spc->status > TGEMPTY) {
                         idx++;
                    }
               }
               else {
                    idx++;
               }
               if (lastidx != idx) {
                    if ((pbuf=prfptr-prfbuf) <= blimit) {
                         setmisc(spr("%d",idx+1));
                         prfmsg(TAGLSTL);
                    }
                    else if (pbuf > blimit && !stoprf) {
                         prfmsg(MORETAGS,ftuptr->numftg-loop);
                         stoprf=1;
                    }
                    numdisp++;
               }
               lastidx=idx;
          }
          if (nlfiles > 1) {
               if (totsize > 0L) {
                    setmisc(fzer(totsize/1024L));
                    prfmsg(DNLSIZE);
                    setmisc(skpwht(dnlmin(totsize)));
                    prfmsg(DNLTIME);
                    lim=usrptr->cltptr->limday;
                    if ((((usaptr->timtdy+15L+totsecs)/60L) >= lim)
                      && (lim != -1)) {
                         warntime=1;
                    }
                    lim=usrptr->cltptr->limcal;
                    if (((usrptr->minut4/4+totsecs/60L+1) >= lim)
                      && (lim != -1)) {
                         warntime=1;
                    }
               }
               if (totchg > 0L) {
                    setmisc(l2as(totchg));
                    prfmsg(DNLCHGS);
                    setmisc(l2as(usaptr->creds+    /* future api */
                       (fndcls(usaptr->curcls)->dbtlmt)));
                    prfmsg(DNLYOURC);
                    if (!tstcrd(totchg)) {
                         prfmsg(numdisp > 1 ? WARNDLM : WARNDLA);
                    }
               }
               if (warntime) {
                    prfmsg(numdisp > 1 ? WARNTMM : WARNTMA);
               }
          }
          if (uwaitcpy(usrnum) == 1) {
               prfmsg(WARNCOPY);
          }
          switch (numdisp) {
          case 0:
               clrprf();
               if (flo->flags&OFTGMAN) {
                    usrptr->substt=PRESENTR;
                    flo->retstt=FILELST0;
                    inpfunc();
                    prf("");
               }
               else {
                    retmenu(0);
               }
               break;
          case 1:
               prfmsg(TAGFOOTA);
               break;
          default:
               prfmsg(TAGFOOTN,lastidx+1,-(lastidx+1));
          }
     }
     return(ftuptr->numftg);
}

int
zapcrd(                            /* check for/stop charging credits      */
int yes,
long siz,
struct fllib *dlib)
{
     int retval=1;

     flo->crdrat=usrptr->crdrat;
     flo->nozap=(int)(usrptr->flags&NOZAP);
     if ((siz > 0L) && (dlib != NULL)) {
          if (!tstcrd(dnlchg(siz,dlib->libname))) {
               retval=0;
          }
     }
     if (yes) {
          usrptr->crdrat=0;
          usrptr->flags|=NOZAP;
     }
     return(retval);
}

struct fllib *
keylib(                            /* function to search with              */
char *libfound,
char *libname,
int *retval)
{
     struct fllib *libptr=NULL;

     libptr=libfind(libfound);
     if (libptr == NULL || libfound[0] == '\0'
        || !strlen(visxlib(libptr,libname))) {
          return(libptr);
     }
     *retval=1;
     if (sameas(libfound,flf->libname)
        && notapped(flf->udate) && !isflop(libptr)) {
          *retval=-1;
     }
     return(libptr);
}

struct fllib *
nxtlib(void)                       /* find next Library in flu's structure */
{
     struct fllib *retval;

     while (flo->index < numoflib) {
          retval=liboff(flo->index++);
          if (valname(retval->libname) && visilib(retval)) {
               return(retval);
          }
     }
     return(NULL);
}

void
optiprf(void)                      /* optimize prfbuf                      */
{
     char *spstart=NULL;

     for (prfptr=prfbuf ; *prfptr != '\0' ; prfptr++) {
          if (spstart == NULL) {
               if (*prfptr == ' ') {
                    spstart=prfptr;
               }
          }
          else if (*prfptr != ' ') {
               if (*prfptr == '\r') {
                    strcpy(spstart,prfptr);
                    prfptr=spstart;
               }
               spstart=NULL;
          }
     }
}

void
bgnfiled(                          /* begin file edit / or just display    */
int display)
{
     char *stgptr;
     int loop;

     setbtv(flfdat);
     setmem(descedit,DESCSIZ,0);
     if (!acqbtv(NULL,&flo->key1,COMPLF)) {
          return;
     }
     strupr(flo->key1.filname);
     stzcpy(flo->fdate,flf->fdate,DATESZ);
     flo->ftim=flf->tim;
     stzcpy(flo->udate,flf->udate,DATESZ);
     flo->siz=flf->siz;
     stzcpy(flo->ulby,flf->ulby,UIDSIZ);
     flo->timeul=flf->utime;
     flo->numdls=atol(flf->numdls);
     stzcpy(descedit,flf->desc,DESCSIZ);
     if (isansiu()) {
          if (!display) {
               fsdrhd(spr("%s",getmsg(EDITINFO)));
               outprf(usrnum);
          }
          fileditp(FILEDITA,(display ? -1 : 1),display);
          stgptr=(display ? getmsg(fsdusr->tmpmsg) : getasc(fsdusr->tmpmsg));
          if (display) {
               tpwipe(stgptr,FFDEXIT,12,3);
               fsddsp(stgptr);
               optiprf();
               outprf(usrnum);
               return;
          }
          tpwipe(stgptr,FFDDLT,6,1);
          tpwipe(stgptr,FFDCHG,3,1);
          fsdbkg(stgptr);
          fsdscb->flddat[FFDNAME].flags|=FFFAVD;
          fsdscb->flddat[FFDLIB].flags|=FFFAVD;
          fsdscb->flddat[FFDDATE].flags|=FFFAVD;
          fsdscb->flddat[FFDTIME].flags|=FFFAVD;
          fsdscb->flddat[FFDSIZE].flags|=FFFAVD;
          fsdscb->flddat[FFDDLT].flags|=FFFAVD;
          fsdscb->flddat[FFDCHG].flags|=FFFAVD;
          if (!isflop(curlib)) {
               fsdscb->flddat[FFDUSER].flags|=FFFAVD;
               fsdscb->flddat[FFDDL].flags|=FFFAVD;
               fsdscb->flddat[FFDULD].flags|=FFFAVD;
               fsdscb->flddat[FFDULT].flags|=FFFAVD;
          }
          if (notapped(flf->udate)) {
               fsdscb->flddat[FFDULD].flags|=FFFAVD;
               fsdscb->flddat[FFDULT].flags|=FFFAVD;
          }
          fsdego(fileditv,fileditd);
          optiprf();
     }
     else {
          if (display) {
               prfmsg(FILEDITX,flo->key1.filname,srt2dat(flo->fdate),
                  flo->key1.libname,nctime(flo->ftim),l2as(flo->siz));
               stgptr="UNAPPROVED";
               if (!notapped(flo->udate)) {
                    stgptr=srt2dat(flo->udate);
               }
               prfmsg(FILEDITY,flo->ulby,l2as(flo->numdls),stgptr,
                    nctime(flo->timeul),dnlmin(flo->siz),
                    l2as(dnlchg(flo->siz,flo->key1.libname)));
               darsdesc(NDESCLN,'\r');
               for (loop=0 ; loop < NDESCLN ; loop++) {
                    if (dargv[loop][0] != '\0') {
                         if (loop == 0) {
                              prfmsg(FILEDITD);
                         }
                         prf("   %s\r",dargv[loop]);
                    }
               }
          }
          else {
               fileditp(FILEDITN,-1,0);
               btuoes(usrnum,1);
               usrptr->substt=FILEDITN;
               prfmsg(FILEDITX,flo->key1.filname,srt2dat(flo->fdate),
                  flo->key1.libname,nctime(flo->ftim),l2as(flo->siz));
               prfmsg(FILEDITZ);
          }
     }
     if (isansiu()) {
          powprf();
     }
     else {
          outprf(usrnum);
     }
     clrprf();
}

int
uwaitcpy(                          /* does this user have files copying?   */
int unum)
{
     int loop,tg=0;
     struct taglib *tagptr;

     for (loop=0 ; loop < systags ; loop++) {
          tagptr=tagoff(loop);
          if (tagptr->usrnum == unum) {
               switch (tagptr->status) {
               case TGINWAIT:
               case TGOTHER:
               case TGSERVER:
               case TGCOPY:
                    return(1);
               default:
                    tg=1;
               case TGEMPTY:
                    break;
               }
          }
     }
     return(tg ? 0 : -1);
}

void
notifu(void)                       /* notify user optionally, assumes ftag */
{
     if (onsysn(uacoff(ftag->usrnum)->userid,1)) {
          if (!(othusp->flags&NOINJO) && !uwaitcpy(othusn)
             && (ntagged(othuap->userid) > 0)) {
               setmbk(flmsg);
               prfmlt(BEEP);
               if (othusp->state == flstt && othusp->substt == PLSWAIT) {
                    othusp->substt=TAGMAN;
               }
               prfmlt(YERFILE);
               rstmbk();
               injoth();
               clrmlt();
          }
     }
     else {
          ftag->status=-ftag->status;
     }
}

int
ntagged(                           /* # of files user has tagged           */
char *uid)
{
     struct taglib *tag;
     int loop,retval=0;

     if (!onsysn(uid,1)) {
          return(0);
     }
     for (loop=0 ; loop < systags ; loop++) {
          tag=tagoff(loop);
          if (tag->status > TGEMPTY && othusn == tag->usrnum) {
               retval++;
          }
     }
     return(retval);
}

void
checkoth(void)                     /* check if others waiting on same file */
{
     struct taglib *test;
     struct libcsfil *ctest;
     int loop;
     char *l,*f;

     l=(ftag == NULL ? lcsarr[usrnum].libname : ftag->libname);
     f=(ftag == NULL ? lcsarr[usrnum].filname : ftag->filname);
     for (loop=0 ; loop < systags ; loop++) {
          test=tagoff(loop);
          if (test != ftag) {
               switch (test->status) {
               case TGINWAIT:
               case TGOTHER:
               case TGSERVER:
                    if (sameas(test->filname,f) && sameas(test->libname,l)) {
                         test->status=TGDLTMP;
                         notifu();
                    }
               }
          }
     }
     for (loop=0 ; loop < nterms ; loop++) {
          ctest=&lcsarr[loop];
          if (ftag != NULL || loop != usrnum) {
               switch (ctest->status) {
               case CDFILE:
               case EXTCDF:
               case EXTCDC:
                    if (sameas(ctest->filname,f) && sameas(ctest->libname,l)) {
                         ctest->status=CDDONE;
                    }
               }
          }
     }
}

void
ulcount(void)                      /* ongoing count of uploads this week   */
{
     int retval=0,ago;
     struct fllib *libptr;

     setbtv(flfdat);
     if (agtbtv(NULL,&countkey,COMPLUF) && sameas(flf->libname,countlib)) {
          stzcpy(countkey.filname,flf->filname,FLFILENM);
          stzcpy(countkey.udate,flf->udate,DATESZ);
          ago=cofdat(today())-cofdat(dcdate(srt2dat(flf->udate)));
          if (ago >= 0 && ago < ULDAYS) {
               countuls[ago]++;
               retval=1;
          }
     }
     if (!retval) {
          libptr=libfind(countlib);
          countlib[0]='\0';
          if (libptr != NULL) {
               for (ago=0 ; ago < ULDAYS ; ago++) {
                    libptr->newfiles[ago]=countuls[ago];
               }
               libptr->flags|=LIBCHN;
          }
     }
     rstbtv();
}

void
inicount(                          /* initialize counting a Library        */
char *libname)
{
     int loop;

     setmem(&countkey,sizeof(struct key0),0);
     stzcpy(countkey.libname,libname,FLNAMESZ);
     stzcpy(countkey.udate,"0000009",DATESZ);
     for (loop=0 ; loop < ULDAYS ; loop++) {
          countuls[loop]=0;
     }
     stzcpy(countlib,libname,FLNAMESZ);
}

int
addkwnow(void)                     /* add the first keyword waiting...     */
{

     if (kwbuf[0].keyword[0] != '\0') {
          setbtv(flkdat);
          if (!acqbtv(NULL,&kwbuf[0],0)) {
               insbtv(&kwbuf[0]);
          }
          rstbtv();
          movmem(&kwbuf[1],&kwbuf[0],(SIZKWBUF-1)*sizeof(struct key7));
          setmem(&kwbuf[SIZKWBUF-1],sizeof(struct key7),0);
          return(1);
     }
     return(0);
}

void
addkw(                             /* add keyword to disk buffer memory    */
char *keyword,
char *filname,
char *libname)
{
     struct key7 *kwp;
     int loop;

     if (samein(spr(" %s ",keyword),nonkwds)) {
          return;
     }
     if ((kwp=&kwbuf[SIZKWBUF-1])->keyword[0] != '\0') {
          addkwnow();
     }
     else {
          for (loop=0 ; loop < SIZKWBUF ; loop++) {
               if ((kwp=&kwbuf[loop])->keyword[0] == '\0') {
                    break;
               }
          }
     }
     stzcpy(kwp->keyword,keyword,FLKEYSIZ);
     stzcpy(kwp->filname,filname,FLFILENM);
     stzcpy(kwp->libname,libname,FLNAMESZ);
}

void
keywdaut(                          /* insert keywords to .DAT one at a time*/
int taskid)
{
     if (!addkwnow()) {
          mfytask(taskid,NULL);
          kwaut=0;
     }
}

int
delkw(
char *keyword,
char *filname,
char *libname)
{
     struct key7 *kwp;
     int loop,retval=0;

     if (libname[0] == '\0') {
          return(0);
     }
     for (loop=0 ; loop < SIZKWBUF ; loop++) {
          if ((kwp=&kwbuf[loop])->libname[0] == '\0') {
               break;
          }
          if (sameas(libname,kwp->libname)) {
               if (keyword[0] == '\0' || sameas(keyword,kwp->keyword)) {
                    if (sameas(filname,kwp->filname)) {
                         movmem(&kwbuf[loop+1],&kwbuf[loop],
                            (SIZKWBUF-(loop+1))*sizeof(struct key7));
                         setmem(&kwbuf[SIZKWBUF-1],sizeof(struct key7),0);
                         return(1);
                    }
               }
          }
     }
     setbtv(flkdat);
     if (filname[0] != '\0') {
          if (acqbtv(NULL,rcompkey(libname,filname),COMPFL_K)) {
               delbtv();
               retval=1;
          }
     }
     rstbtv();
     return(retval);
}

void
copyaut(                           /* copy a file, in chunks               */
int taskid)
{
     static long sleft;
     int readsiz,close=0,csu;
     char *stg,dpath[80];
     struct fllib *libptr=NULL;
     struct fndblk fb;

     if ((csu=findbest(TGCOPY)) == 0) {
          mfytask(taskid,NULL);
          return;
     }
     if (!space()) {
          return;
     }
     if (ftag == NULL) { /* therefore it's a c/s tag */
          csu--;
          if (csrc == NULL && cdst == NULL) {
               libptr=libfind(lcsarr[csu].libname);
               if (libptr == NULL) {
                    lcsarr[csu].status=-lcsarr[csu].status;
                    return;
               }
               stg=spr("%s\\%s",libpath(libptr),lcsarr[csu].filname);
               readsiz=fnd1st(&fb,stg,0);
               sprintf(dpath,"%s\\%s",copydir,libptr->libname);
               mkdir(dpath);
               strcat(dpath,"\\");
               strcat(dpath,lcsarr[csu].filname);
               cdst=fopen(dpath,FOPWB);
               if (cdst != NULL) {
                    if (readsiz && (csrc=fopen(stg,FOPRB)) != NULL) {
                         sleft=fb.size;
                    }
                    else {
                         lcsarr[csu].status=-lcsarr[csu].status;
                         fclose(cdst);
                         cdst=NULL;
                         unlink(dpath);
                    }
               }
               else {
                    lcsarr[csu].status=-lcsarr[csu].status;
               }
               return;
          }
          readsiz=1024;
          if (sleft < 1024L) {
               readsiz=(int)sleft;
          }
          sleft-=(long)readsiz;
          if ((int)fread(vdatmp,readsiz,1,csrc) <= 0
           || (int)fwrite(vdatmp,readsiz,1,cdst) <= 0) {
               lcsarr[csu].status=-lcsarr[csu].status;
               close=1;
          }
          if (close || sleft == 0L) {
               fclose(csrc);
               csrc=NULL;
               fclose(cdst);
               cdst=NULL;
               libptr=libfind(lcsarr[csu].libname);
               if (libptr != NULL) {
                    stg=spr("%s\\%s",libpath(libptr),lcsarr[csu].filname);
                    sprintf(dpath,"%s\\%s",copydir,libptr->libname);
                    mkdir(dpath);
                    strcat(dpath,"\\");
                    strcat(dpath,lcsarr[csu].filname);
                    touch(stg,dpath);
               }
               if (!close) {
                    lcsarr[csu].status=CDDONE;
                    usrnum=csu;
                    checkoth();
               }
          }
          return;
     }
     if (csrc == NULL && cdst == NULL) {
          libptr=libfind(ftag->libname);
          if (libptr == NULL) {
               ftag->status=-ftag->status;
               return;
          }
          stg=spr("%s\\%s",libpath(libptr),ftag->filname);
          readsiz=fnd1st(&fb,stg,0);
          sprintf(dpath,"%s\\%s",copydir,ftag->libname);
          mkdir(dpath);
          strcat(dpath,"\\");
          strcat(dpath,ftag->filname);
          cdst=fopen(dpath,FOPWB);
          if (cdst != NULL) {
               if (readsiz && (csrc=fopen(stg,FOPRB)) != NULL) {
                    ftag->size=sleft=fb.size;
               }
               else {
                    ftag->status=-ftag->status;
                    fclose(cdst);
                    cdst=NULL;
                    unlink(dpath);
               }
          }
          else {
               ftag->status=-ftag->status;
          }
          return;
     }
     readsiz=1024;
     if (sleft < 1024L) {
          readsiz=(int)sleft;
     }
     sleft-=(long)readsiz;
     ftag->copied+=readsiz;
     if ((int)fread(vdatmp,readsiz,1,csrc) <= 0
      || (int)fwrite(vdatmp,readsiz,1,cdst) <= 0) {
          ftag->status=-ftag->status;
          close=1;
     }
     if (close || sleft == 0L) {
          fclose(csrc);
          csrc=NULL;
          fclose(cdst);
          cdst=NULL;
          libptr=libfind(ftag->libname);
          if (libptr != NULL) {
               stg=spr("%s\\%s",libpath(libptr),ftag->filname);
               sprintf(dpath,"%s\\%s",copydir,ftag->libname);
               mkdir(dpath);
               strcat(dpath,"\\");
               strcat(dpath,ftag->filname);
               touch(stg,dpath);
          }
          if (!close) {
               ftag->status=TGDLTMP;
               notifu();
               checkoth();
          }
     }
}

void
doalr(void)                        /* re-mark files already tagged         */
{
     int loop;

     for (loop=0 ; loop < NLISTER ; loop++) {
          if (alrtag(flo->tags[loop].libname,flo->tags[loop].filname)) {
               if (flo->tdline[loop] > 0) {
                    flo->tdline[loop]=-flo->tdline[loop];
               }
          }
     }
}

void
dsearch(                           /* recursively get how user searches    */
int which,
int x)
{
     int mode;

     switch (which) {
     case 0:
          wipetag();
          deresume();
          setmem(flo->keylist,FLKEYLST,0);
          if (x == 0) {
               x=defstyp;
          }
          switch (x) {
          default:
          case 1:
               usrptr->substt=FLSTYPA;
               break;
          case 7:
               flo->styp=2;
               flo->flags|=DATESRC;
               dsearch(1,((flo->flags&ABOUPRF) ? 1 : flu->mylib));
               break;
          case 2:
               flo->flags&=~DATESRC;
          case 5:
               flo->styp=x;
               dsearch(1,((flo->flags&ABOUPRF) ? 1 : flu->mylib));
               break;
          case 3:
          case 6:
               flo->styp=(x == 3 ? 3 : 7);
               usrptr->substt=ASKNAME;
               break;
          case 4:
               flo->styp=4;
               usrptr->substt=ASKWORD;
               btumil(usrnum,FLKEYLST-1);
          }
          break;
     case 1:
          if (x == 0) {
               x=defslib;
          }
          switch (x) {
          default:
          case 1:
               usrptr->substt=FLSLIB;
               break;
          case 2:
          case 3:
               usrptr->substt=MMENU;
               switch (flo->styp) {
               case 3:
               case 4:
                    mode=flo->styp-1;
                    break;
               case 6:
                    mode=4;
                    break;
               case 5:
                    mode=5;
                    break;
               case 7:
                    mode=6;
                    break;
               default:
                    mode=1;
               }
               srcinit(mode,(x == 2) ? flu->lib : "");
          }
     }
}

void
fileditd(                          /* FSD finish editing file desc/etc.    */
int save)
{
     int loop,len,tim;
     char temp[UIDSIZ],date[DATESZ],*sptr;

     flu=&fluarr[usrnum];
     if (flu->lib[0] == '\0') {
          return;
     }
     setmbk(flmsg);
     curusr(usrnum);
     fluoff(usrnum);
     usrptr->state=flstt;
     usrptr->substt=flo->retstt;
     if ((fsdscb->chgcnt == 0) && (libfind(libedt->lib.libname) != NULL)) {
          prfmsg(NOCH);
     }
     else if (!save) {
          prfmsg(CHQUIT);
     }
     else {
          prfmsg(SAVED);
          setmem(descedit,DESCSIZ,0);
          for (loop=0 ; loop < NDESCLN ; loop++) {
               len=strlen(descedit);
               fsdfxt(FFDTXT00+loop,sptr=&descedit[strlen(descedit)],SDESCLN);
               if (!loop && sameas(sptr,DESCFNON) ||
                  loop == 1 && sameas(sptr,DESCFNO2)) {
                    *sptr='\0';
               }
               if (strlen(descedit) > len) {
                    strcat(descedit,"\r");
               }
          }
          fsdfxt(FFDUSER,flo->ulby,UIDSIZ);
          fsdfxt(FFDDL,temp,NUMSZ);
          flo->numdls=atol(temp);
          fsdfxt(FFDULD,temp,UIDSIZ);
          stzcpy(date,dat2srt(temp),DATESZ);
          fsdfxt(FFDULT,temp,UIDSIZ);
          tim=dctime(temp);
          setbtv(flfdat);
          if (acqbtv(NULL,&flo->key1,COMPLF)) {
               stzcpy(flf->ulby,flo->ulby,UIDSIZ);
               sprintf(flf->numdls,NUMSZP,flo->numdls);
               stzcpy(flf->desc,descedit,DESCSIZ);
               stzcpy(flf->udate,(atoi(date) != 0) ? date : NOTAPPED,DATESZ);
               flf->utime=tim;
               upvbtv(NULL,FLFILREC+strlen(flf->desc)+1);
               flo->retstt=usrptr->substt;
               if (!longsrch && kwdout()) {
                    outprf(usrnum);
                    clrprf();
                    return;
               }
               usrptr->substt=flo->retstt;
          }
     }
     outprf(usrnum);
     if (flo->retstt == AUTOLOG) {
          usrptr->substt=AUTOLOG;
          btuinj(usrnum,CYCLE);
     }
     else {
          btucli(usrnum);
          fluoff(usrnum);
          promptin();
     }
     clrprf();
}

void
fileditp(                          /* FSD utility function                 */
int msgno,
int amode,
int display)
{
     char *appdate,*proprlib;
     struct fllib *libptr;

     libptr=libfind(flo->key1.libname);
     proprlib=((display && libptr != NULL) ? visxlib(libptr,flo->lrange)
        : flo->key1.libname);
     appdate=(notapped(flo->udate) ? NULL : srt2dat(flo->udate));
     fsdroom(msgno,filedits,amode);
     darsdesc(NDESCLN,'\r');
     sprintf(vdatmp,fileditf,
          flo->key1.filname,'\0',
          proprlib,'\0',
          srt2dat(flo->fdate),'\0',
          spr("%s",nctime(flo->ftim)),'\0',
          spr("%s bytes",l2as(flo->siz)),'\0',
          l2as(flo->numdls),'\0',
          (appdate == NULL ? "UNAPP" : appdate),'\0',
          spr("%s",nctime(flo->timeul)),'\0',
          display ? spr("%s [32mminutes",dnlmin(flo->siz)) : "",'\0',
          display ? l2as(dnlchg(flo->siz,flo->key1.libname)) : "",'\0',
          flo->ulby,'\0',
          (display || strlen(dargv[0]) ? dargv[0] : DESCFNON),'\0',
          (display || strlen(dargv[1]) ? dargv[1] : DESCFNO2),'\0',
          dargv[2],'\0',
          dargv[3],'\0',
          dargv[4],'\0',
          dargv[5],'\0',
          dargv[6],'\0',
          dargv[7],'\0',
          dargv[8],'\0',
          dargv[9],'\0',
          dargv[10],'\0',
          dargv[11],'\0',
          dargv[12],'\0',
          dargv[13],'\0',
          dargv[14],'\0');
     fsdapr(vdaptr,vdasiz,vdatmp);
}

void
lsclean(void)                      /* nightly cleanup routine              */
{
     switch (clnlevel) {
     case 2:
     case 1:
          countupl();
          cleanit(clnlevel == 2 ? 1 : 0);
     default:
          do {
          } while (libchg());
     }
     purgecpy(1);
}

void
lsdel(                             /* deletes Library user reference       */
char *uid)
{
     setbtv(genbb);
     stzcpy(flusmem.userid,uid,UIDSIZ);
     stzcpy(flusmem.modnam,"File Library",MNMSIZ);
     if (acqbtv(NULL,&flusmem,0)) {
          delbtv();
     }
}

void
lsend(void)                        /* system shutdown                      */
{
     while (libchg()) {
     }
     if (kwaut) {
          while (addkwnow()) {
          }
          kwaut=0;
     }
     if (flmsg != NULL) {
          clsmsg(flmsg);
     }
     if (csrc != NULL) {
          fclose(csrc);
     }
     if (cdst != NULL) {
          fclose(cdst);
     }
     if (movs != NULL) {
          fclose(movs);
     }
     if (movd != NULL) {
          fclose(movd);
     }
     clsbtv(flfdat);
     clsbtv(flldat);
     clsbtv(flkdat);
     purgecpy(0);
}

void
purgecpy(                          /* purge copy and/or upload directories */
int rmvdirs)
{
     struct fndblk dirs,fils;

     if (fnd1st(&dirs,spr("%s\\*.*",copydir),FAMDIR)) {
          do {
               if ((dirs.attr&FAMDIR) && dirs.name[0] != '.') {
                    if (fnd1st(&fils,spr("%s\\%s\\*.*",copydir,dirs.name),0)) {
                         do {
                              unlink(spr("%s\\%s\\%s",
                                   copydir,dirs.name,fils.name));
                         } while (fndnxt(&fils));
                    }
                    if (rmvdirs) {
                         rmdir(spr("%s\\%s",copydir,dirs.name));
                    }
               }
          } while (fndnxt(&dirs));
     }
     if (rmvdirs && fnd1st(&fils,spr("%s\\*.*",userdir(0)),0)) {
          do {
               unlink(spr("%s\\%s",userdir(0),fils.name));
          } while (fndnxt(&fils));
     }
}

void
lshang(void)                       /* user hangup routine                  */
{
     char fname[FLFILENM];
     int ctr=0;

     if (usrptr->state == flstt) {
          echonu(usrnum);
          btutrg(usrnum,0);
     }
     genhup();
     if (flo->flags&AXSUSER) {
          while (ulfileq(2) && ctr < 5) {
               ctr++;
          }
          rmdir(userdir(1));
          sprintf(fname,"*.x%02X",channel[usrnum]);
          if (fnd1st(&flo->fb,spr("%s\\%s",copydir,fname),0)) {
               do {
                    unlink(spr("%s\\%s",copydir,flo->fb.name));
               } while (fndnxt(&flo->fb));
          }
     }
}

void
genhup(void)                       /* C/S - A/A generic hangup routine     */
{
     fluoff(usrnum);
     if (lcsarr[usrnum].status != 0) {
          lcsarr[usrnum].status=-abs(lcsarr[usrnum].status);
          csclean(usrnum);
     }
     setmem(&viewarr[usrnum],sizeof(struct libcsfil),0);
     if (flo->cszview.state != 0) {
          fclose(flo->cszview.fp);
          flo->cszview.state=0;
     }
     if (flo->fsrc != NULL) {
          fclose(flo->fsrc);
     }
     if (flo->fdst != NULL) {
          fclose(flo->fdst);
     }
     flo->fdst=flo->fsrc=NULL;
     donewrit(-1);
     updusr(flu);
     flu->lib[0]='\0';
}

